/*
 * Test program for USB-Tiny-CAN
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 * 
 *
 * Description: 
 *		Test program for USB-Tiny-CAN
 *
 * TODO:
 *
 * History:
 *	 2005/03/15 - 0.1 - first version
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "stdafx.h"
#include <winioctl.h>
#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include "..\tcan_usb\ioctls.h"
#include "..\tcan_usb\tcan_usb_global.h"

void PrintUsage ()
{
	printf ( "-------------------------------------------------------\n" );
	printf ( "USB-Tiny-CAN Test Tool\n" );
	printf ( "-------------------------------------------------------\n" );
	printf ( "Author: Chi-Tai, Dang\n" );
	printf ( "        thesis at chi-tai dot info\n" );
	printf ( "*******************************************************\n" );
	printf ( "usage: tcan_test.exe DeviceNumber FunctionCode Argument\n" );
	printf ( " e.g.: tcan_test.exe 0 24 500\n" );
	printf ( "*******************************************************\n" );
	printf ( "Function Code: Description\n" );
	printf ( "0: get manufacturer string - IOCTL_GET_MANUFACTURER\n" );
	printf ( "1: get product string - IOCTL_GET_PRODUCT\n" );
	printf ( "2: get serialnumber - IOCTL_GET_SERIALNUMBER\n" );
	printf ( "3: get driver version - IOCTL_DRIVER_VERSION\n" );
	printf ( "5: testing set baudrate - IOCTL_SET_BAUD\n" );
	printf ( "6: can stop - IOCTL_STOP\n" );
	printf ( "7: can reset - IOCTL_RESET\n" );
	printf ( "8: set acc code 0x12345678 - IOCTL_SET_ACC_CODE\n" );
	printf ( "9: get acc code - IOCTL_GET_ACC_CODE\n" );
	printf ( "10: set acc code 0x00000000 - IOCTL_SET_ACC_CODE\n" );
	printf ( "11: get acc code - IOCTL_GET_ACC_CODE\n" );
	printf ( "12: set acc mask 0x12345678 - IOCTL_SET_ACC_MASK\n" );
	//printf ( "13: get acc mask - IOCTL_GET_ACC_MASK\n" );
	printf ( "14: set acc mask 0xffffffff - IOCTL_SET_ACC_MASK\n" );
	//printf ( "15: get acc mask - IOCTL_GET_ACC_MASK\n" );
	printf ( "16: can start - IOCTL_START\n" );
	printf ( "17: get interrupt count - IOCTL_GET_INTERRUPTCOUNT\n" );
	printf ( "18: send 400 messages\n" );
	printf ( "19: loop send of 400 messages\n" );
	printf ( "22: read bulk in\n" );
	printf ( "23: set io mode\n" );
	printf ( "24: create random bulk data, Argument = max Sleeptime\n" );
}
  
int main ( int argc, char* argv[] )
{
	UINT DevNum = 0;
	UINT FCode = 18;
	UINT Argument = 1000;
	if ( argc > 2 )
	{
		DevNum = (UINT) _atoi64 ( argv[1] );
		FCode = (UINT) _atoi64 ( argv[2] );
		if ( argc > 3 ) Argument = (UINT) _atoi64 ( argv[3] );
		if ( FCode > 100 ) FCode = 24;
	}
	/*if ( argc < 3 )
	{
		PrintUsage ();
		return 0;
	}*/
	//printf ( "Arguments: %i %i %i\n", DevNum, FCode, Argument );
	
	char DevName[255];
	sprintf ( DevName, "\\\\.\\CANUSB%i", DevNum );
	//printf ( "Device: %s\n", DevName );

	//HANDLE hdevice = CreateFile ( "\\\\.\\CANUSB0", GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
	HANDLE hdevice = CreateFile ( DevName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
	if ( hdevice == INVALID_HANDLE_VALUE )
	{
		printf ( "Failed to open CANUSB0 device - error %d\n",  GetLastError () );
		return 1;
	}
	CHAR hline[] = "*****************************************************************\n";
	BYTE retval[64];
	DWORD foo;
	DWORD data = 0x00000000;
	switch ( FCode )
	{
		case 0:
		{
			HANDLE hEvent = CreateEvent ( NULL, FALSE, FALSE, NULL );
			/****************************
				get manufacturer string
			****************************/
			printf ( "%sTesting: IOCTL_GET_MANUFACTURER\n", hline);
			printf ( "Description: urb to read in manufacturer string descriptor\n%s", hline);
			if ( DeviceIoControl ( hdevice, IOCTL_GET_MANUFACTURER, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: Manufacturer %ws\n",  retval );
			}
			else
				printf ( "!FAILED! with %d\n",  GetLastError () );
			break;
		}
		case 1:
		{
			/****************************
				get product string
			****************************/
			printf ( "\n\n%sTesting: IOCTL_GET_PRODUCT\n", hline );
			printf ( "Description: urb to read in product string descriptor\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_GET_PRODUCT, NULL, 0, retval, sizeof(retval), &foo, NULL))
			{
				printf ( "SUCCESS: Product %ws\n",  retval );
			}
			else
				printf ( "!FAILED! with %d\n", GetLastError () );
			break;
		}
		case 2:
		{
			/****************************
				get serialnumber
			****************************/
			printf ( "\n\n%sTesting: IOCTL_GET_SERIALNUMBER\n", hline );
			printf ( "Description: urb to read in serialnumber string descriptor\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_GET_SERIALNUMBER, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: Serialnumber %ws\n",  retval );
			}
			else
				printf ( "!FAILED! with %d\n", GetLastError () );
			break;
		}
		case 3:
		{
			/****************************
				get driver version
			****************************/
			printf ( "\n\n%sTesting: IOCTL_DRIVER_VERSION\n", hline );
			printf ( "Description: get driver version\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_DRIVER_VERSION, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: Version of CANUSB0 is %.2x%.2x.%.2x%.2x\n",  retval[3], retval[2], retval[1], retval[0] );
			}
			else
				printf ( "!FAILED! with %d\n", GetLastError () );
			break;
		}
		case 4:
		{
			/****************************
				avr status
			****************************/
			printf ( "\n\n%sTesting: IOCTL_GET_AVR_STATUS\n", hline );
			printf ( "Description: urb to get avr status\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_GET_AVR_STATUS, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: got avr status %X\n",  retval[0] );
			}
			else
				printf ( "!FAILED! get avr status %X and LastError %d\n",  retval[0], GetLastError () );
			break;
		}
		case 5:
		{
			/****************************
				testing set baudrate
			****************************/
			printf ( "\n\n%sTesting: IOCTL_SET_BAUD\n", hline );
			printf ( "Description: control request urb to set baud rates\n%s", hline );
			
			for ( unsigned int i=0; i <= CAN_BAUD_ENDVALUE; i++ )
			{
				data = i;
				if ( DeviceIoControl ( hdevice, IOCTL_SET_BAUD, &data, 1, retval, sizeof(retval), &foo, NULL ) )
				{
					printf ( "SUCCESS: set to baudrate with index %i\n", i );
				}
				else
					printf ( "!FAILED! set to baudrate %i with %d\n", i, GetLastError () );
			}
			break;
		}
		case 6:
		{
			/****************************
				can stop
			****************************/
			printf ( "\n\n%sTesting: IOCTL_STOP\n", hline );
			printf ( "Description: urb to stop can\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_STOP, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: can stop\n" );
			}
			else
				printf ( "!FAILED! can stop with LastError %d\n", GetLastError () );
			break;
		}
		case 7:
		{
			/****************************
				can reset
			****************************/
			printf ( "\n\n%sTesting: IOCTL_RESET\n", hline );
			printf ( "Description: urb to reset can\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_RESET, NULL, 0, NULL, 0, &foo, NULL ) )
			{
				printf ( "SUCCESS: can reset\n" );
			}
			else
				printf ( "!FAILED! can reset with LastError %d\n", GetLastError () );
			break;
		}
		case 8:
		{
			/****************************
				set acc code
			****************************/
			printf ( "\n\n%sTesting: IOCTL_SET_ACC_CODE\n", hline );
			printf ( "Description: urb to set acc code 0x12345678\n%s", hline );
			data = 0x12345678;
			if ( DeviceIoControl ( hdevice, IOCTL_SET_ACC_CODE, &data, 4, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: set acc code %.8x\n", data );
			}
			else
				printf ( "!FAILED! set acc code with %d\n", GetLastError () );
			break;
		}
		case 9:
		{
			/****************************
				get can acc code
			****************************/
			printf ( "\n\n%sTesting: IOCTL_GET_ACC_CODE\n", hline );
			printf ( "Description: urb to get acc code\n%s", hline );
			retval[0] = 0;retval[1] = 0;retval[2] = 0;retval[3] = 0;
			if ( DeviceIoControl ( hdevice, IOCTL_GET_ACC_CODE, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: get can acc code %.2x %.2x %.2x %.2x \n", retval[0], retval[1], retval[2], retval[3] );
			}
			else
				printf ( "!FAILED! get can acc code with %X and LastError %d\n", (DWORD) *retval, GetLastError () );
			break;
		}
		case 10:
		{
			/****************************
				set acc code
			****************************/
			printf ( "\n\n%sTesting: IOCTL_SET_ACC_CODE\n", hline );
			printf ( "Description: urb to set acc code 0x00000000\n%s", hline );
			data = 0x00000000;
			if ( DeviceIoControl ( hdevice, IOCTL_SET_ACC_CODE, &data, 4, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: set acc code %.8x\n", data );
			}
			else
				printf ( "!FAILED! set acc code with %d\n", GetLastError () );
			break;
		}
		case 11:
		{
			/****************************
				get can acc code
			****************************/
			printf ( "\n\n%sTesting: IOCTL_GET_ACC_CODE\n", hline );
			printf ( "Description: urb to get acc code\n%s", hline );
			retval[0] = 0;retval[1] = 0;retval[2] = 0;retval[3] = 0;
			if ( DeviceIoControl ( hdevice, IOCTL_GET_ACC_CODE, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: get can acc code %.2x %.2x %.2x %.2x \n",  retval[0], retval[1], retval[2], retval[3] );
			}
			else
				printf ( "!FAILED! get can acc code with %X and LastError %d\n", (DWORD) *retval, GetLastError () );
			break;
		}
		case 12:
		{
			/****************************
				set acc mask
			****************************/
			printf ( "\n\n%sTesting: IOCTL_SET_ACC_MASK\n", hline );
			printf ( "Description: urb to set acc mask 0x12345678\n%s", hline );
			data = 0x12345678;
			if ( DeviceIoControl ( hdevice, IOCTL_SET_ACC_MASK, &data, 4, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: set acc mask %.8x\n",  data );
			}
			else
				printf ( "!FAILED! set acc mask with %d\n", GetLastError () );
			break;
		}
		case 13:
		{
			/****************************
				...
			****************************/
			break;
		}
		case 14:
		{
			/****************************
				set acc mask
			****************************/
			data = 0xffffffff;
			printf ( "\n\n%sTesting: IOCTL_SET_ACC_MASK\n", hline );
			printf ( "Description: urb to set acc mask %X\n%s",data,hline );
			if ( DeviceIoControl ( hdevice, IOCTL_SET_ACC_MASK, &data, 4, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: set acc mask %.8x\n",  data );
			}
			else
				printf ( "!FAILED! set acc mask with %d\n", GetLastError () );
			break;
		}
		case 15:
		{
			/****************************
				...
			****************************/
			break;
		}
		case 16:
		{
			/****************************
				can start
			****************************/
			printf ( "\n\n%sTesting: IOCTL_START\n", hline );
			printf ( "Description: urb to start can\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_START, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: can start with %d\n",  retval[0] );
			}
			else
				printf ( "!FAILED! can start with %d and LastError %d\n", retval[0], GetLastError () );
			break;
		}
		case 17:
		{
			/****************************
				get interrupt count
			****************************/
			printf ( "\n\n%sTesting: IOCTL_GET_INTERRUPTCOUNT\n", hline );
			printf ( "Description: get interrupt count\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_GET_INTERRUPTCOUNT, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS: count of interrupts %X\n", *((PULONG) retval) );
			}
			else
				printf ( "!FAILED! with %d\n", GetLastError () );
			break;
		}
		case 18:
		{
			/****************************
				send 400 messages
			****************************/
			CANMSG msg[400];
			memset(msg, 0, sizeof(msg) );
			
			sprintf ( (char *) msg[0].msgdata, "c_msg0" ); // default msg "c_msg0"
			msg[0].flags = 6; // default msg length 6
			msg[0].id = 0; // default msg id 0 

			int count = 400;
			int len = (int) strlen ( (char *) msg[0].msgdata );
			for ( int i = 1; i <= (count-1); i++ ) {
				sprintf ( (PCHAR)msg[i].msgdata, (PCHAR)msg[i-1].msgdata );
				msg[i].msgdata[len-1]++;
				msg[i].flags = (0x0f & len) % 9;
				msg[i].flags |= CAN_EFF;
				msg[i].id = i;
			}
			
			if ( WriteFile ( hdevice, msg, 400 * sizeof(CANMSG), &foo, NULL) )
				printf ( "Write transferred %d bytes okay\n", foo );
			else
				printf ( "Error %d trying to write data\n", GetLastError () ); 
			break;
		}
		case 19:
		{
			/****************************
				loop send of 400 messages
			****************************/
			CANMSG msg[400];
			CloseHandle ( hdevice );
			memset(msg, 0, sizeof(msg) );
			
			sprintf ( (char *) msg[0].msgdata, "c_msg0" ); // default msg "c_msg0"
			msg[0].flags = 6; // default msg length 6
			msg[0].id = 0; // default msg id 0 

			int count = 400;
			int len = (int) strlen ( (char *) msg[0].msgdata );
			for ( int i = 1; i <= (count-1); i++ ) {
				sprintf ( (PCHAR)msg[i].msgdata, (PCHAR)msg[i-1].msgdata );
				msg[i].msgdata[len-1]++;
				msg[i].flags = (0x0f & len) % 9;
				msg[i].flags |= CAN_EFF;
				msg[i].id = i;
			}
			while ( 1 )
			{
				HANDLE hLdevice = CreateFile ( DevName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
				if ( hdevice == INVALID_HANDLE_VALUE )
				{
					printf ( "Failed to open CANUSB%i device - error %d\n", DevNum,  GetLastError () );
					continue;
				}
				if ( WriteFile ( hdevice, msg, 400 * sizeof(CANMSG), &foo, NULL) )
					printf ( "Write transferred %d bytes okay\n", foo );
				else
					printf ( "Error %d trying to write data\n", GetLastError () );
				CloseHandle ( hLdevice );
			}
			break;
		}
		case 20:
		{
			/****************************
				send 1 message
			****************************/
			CANMSG msg;
			memset(&msg, 0, sizeof(CANMSG) );
			
			sprintf ( (char *) &msg.msgdata, "c_msg0" ); // default msg "c_msg0"
			msg.flags = 6; // default msg length 6
			msg.id = 0; // default msg id 0
			msg.flags |= CAN_EFF; 
			int len = (int) strlen ( (char *) msg.msgdata );
			msg.flags = (0x0f & len) % 9;

			msg.id = rand() % 400;
			if ( WriteFile ( hdevice, &msg, sizeof(CANMSG), &foo, NULL) )
				printf ( "Write transferred %d bytes okay\n", foo );
			else
				printf ( "ERROR %d trying to write data\n", GetLastError () );

			break;
		}
		/*case 19:
		{
			***************************
				write bulk out
			***************************
			printf ( "\n\n%sTesting: IOCTL_WRITE_SIMPLE\n", hline );
			printf ( "Description: urb to write bulk out\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_WRITE_SIMPLE, msg, sizeof(CANMSG)*count, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS:\nwritten: %X\n", foo );
			}
			else
				printf ( "!FAILED! with %d\n", GetLastError () );
			break;
		}*/
		/*case 20:
		{
			***************************
				read bulk in
			***************************
			printf ( "\n\n%sTesting: IOCTL_READ\n", hline );
			printf ( "Description: urb to read bulk in\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_READ, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS:\nData:\n" );
				for (int i=0; i<foo; i++)
					printf ( "%.2x", retval[i] );
				printf ( "\n" );
			}
			else
				printf ( "!FAILED! with %d\n", GetLastError () );
			break;
		}*/
		/*case 21:
		{
			***************************
				loop read bulk in
			***************************
			printf ( "\n\n%sTesting: IOCTL_READ_KDPRINT\n", hline );
			printf ( "Description: loop urb to read bulk in\n%s", hline );
			if ( DeviceIoControl ( hdevice, IOCTL_READ_KDPRINT, NULL, 0, retval, sizeof(retval), &foo, NULL ) )
				printf ( "SUCCESS:\nData:\n" );
			else
				printf ( "!FAILED! with %d\n", GetLastError () );
			break;
		}*/
		case 22:
		{
			/****************************
				read bulk in
			****************************/
			printf ( "\n\n%sTesting: IOCTL_GET_INTERRUPT\n", hline );
			printf ( "Description: urb to read bulk in\n%s", hline );
			OVERLAPPED ovl;
			ovl.hEvent = CreateEvent (NULL, TRUE, FALSE, NULL );

			if ( DeviceIoControl ( hdevice, IOCTL_GET_INTERRUPT, NULL, 0, retval, sizeof(retval), &foo, &ovl ) )
			{
				printf ( "SUCCESS:\nData:\n" );
				for ( unsigned int i=0; i<foo; i++ )
					printf ( "%.2x", retval[i] );
				printf ( "\n" );
			}
			else
				printf ( "!FAILED! with %d\n", GetLastError () );

			if ( GetOverlappedResult ( hdevice, &ovl, &foo, TRUE ) )
			{
				printf ( "SUCCESS:\nData:\n" );
				for ( unsigned int i=0; i<foo; i++ )
					printf ( "%.2x", retval[i] );
				printf ( "\n" );
			}
			else
				printf ( "!FAILED! with %d\n", GetLastError () );
			break;
		}
		case 23:
		{
			/****************************
				set io mode
			****************************/
			printf ( "\n\n%sTesting: IOCTL_SET_IO_MODE\n", hline );
			printf ( "Description: set io mode to %s\n%s", ( Argument ? "PollMode" : "EventMode" ), hline );
			CIOMODE c = ( Argument ? PollMode : EventMode );
			if ( DeviceIoControl ( hdevice, IOCTL_SET_IO_MODE, &c, sizeof(CIOMODE), retval, sizeof(retval), &foo, NULL ) )
			{
				printf ( "SUCCESS:\nwritten: %X\n", foo );
			}
			else
				printf ( "!FAILED! with %d\n", GetLastError () );
			break;
		}
		case 24:
		{
			/****************************
				create random bulk data
			****************************/
			CANMSG msg;
			UINT	cnt = 0;
			CloseHandle ( hdevice );
			memset(&msg, 0, sizeof(CANMSG) );
			
			sprintf ( (char *) &msg.msgdata, "c_msg0" ); // default msg "c_msg0"
			msg.flags = 6; // default msg length 6
			msg.id = 0; // default msg id 0
			msg.flags |= CAN_EFF; 
			int len = (int) strlen ( (char *) msg.msgdata );
			msg.flags = (0x0f & len) % 9;
            
			Argument = ( Argument ? Argument : 10 );
			while ( 1 )
			{
				HANDLE hLdevice = CreateFile ( DevName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL );
				if ( hdevice == INVALID_HANDLE_VALUE )
				{
					printf ( "Failed to open CANUSB%i device - error %d\n", DevNum,  GetLastError () );
					continue;
				}
				msg.id = rand() % 400;	
				printf ( "%i:\tWrite message...", cnt ); 
				if ( WriteFile ( hdevice, &msg, sizeof(CANMSG), &foo, NULL) )
					printf ( "Write transferred %d bytes okay\n", foo );
				else
					printf ( "ERROR %d trying to write data\n", GetLastError () );
				CloseHandle ( hLdevice );
				Sleep ( rand() % Argument );
				cnt++;
			}
			break;
		}
		default:
		{
			PrintUsage ();
			printf ( "Unknown Function Code (%i) !!! \n", FCode );
			break;
		}
	}
	printf ( " .... \n" );
	CloseHandle ( hdevice );
	scanf ( "bitte return druecken...", foo );
	return 0;
}

