/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description:
 *		Interrupt routines for tcan_usb driver
 *
 * TODO:
 *
 * History of changes:
 *	 2005/03/15 - 0.1 - first version
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "tcan_usb.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text( PAGE, Tc_LoadDevice)
#endif 

/******************************************************************************
	
	  Tc_LoadDevice

 Description:
	This is where all the action is done to set set the programm of tiny CAN
	with reset, download and run the firmware to wilfully 
	expect a surprise removal and hence a remove
	device irp to out pnp code

  @deviceObject - the device object

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_LoadDevice
(
 IN PDEVICE_OBJECT	deviceObject
)
{
	URB urb;
	PCHAR buff;
	HANDLE hFile;
	UCHAR buffer;
	UINT FWLength;
	NTSTATUS ntStatus;
	UNICODE_STRING fileName;
	IO_STATUS_BLOCK IoStatus;
	PDEVICE_EXTENSION devExt;
	OBJECT_ATTRIBUTES objAtt;
	FILE_STANDARD_INFORMATION fsi;
	UINT written;
	ULONG request;

	KdPrint (( DRVNAME "Tc_LoadDevice\n" ));
	PAGED_CODE ();

	devExt 
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	RtlInitUnicodeString(&fileName, L"\\SystemRoot\\system32\\drivers\\tcan_usb.bin");

	InitializeObjectAttributes (
		&objAtt, 
		&fileName, 
		OBJ_CASE_INSENSITIVE | OBJ_KERNEL_HANDLE, 
		NULL, 
		NULL);
	

	ntStatus = ZwCreateFile (
		&hFile, 
		GENERIC_READ, 
		&objAtt, 
		&IoStatus, 
		NULL,
		0, 
		FILE_SHARE_READ, 
		FILE_OPEN, 
		FILE_SYNCHRONOUS_IO_NONALERT, 
		NULL, 
		0);

	if ( !NT_SUCCESS(ntStatus) ) 
	{
		KdPrint (( DRVNAME "FAILED to load tcan_usb.*\n" ));
		return ntStatus;
	}
	
	/*
	   get file information
	*/

	ntStatus
		= ZwQueryInformationFile ( hFile, &IoStatus, (PVOID) &fsi, sizeof(fsi), FileStandardInformation );

	devExt->FWLength = (ULONG) ( NT_SUCCESS(ntStatus) ? fsi.EndOfFile.QuadPart : 0 );

	if ( !devExt->FWLength )
	{
		KdPrint (( DRVNAME "FAILED; tcan_usb.hex with 0 bytes\n" ));
		return ntStatus;
	}
	
	/*
	   Allocate memory and load the file
	*/
	devExt->FWBuffer = (PCHAR) ExAllocatePool(NonPagedPool, devExt->FWLength);

	if ( !devExt->FWBuffer )
	{
		KdPrint (( DRVNAME "FAILED to allocate memory\n" ));
		return STATUS_NO_MEMORY;
	}

	ZwReadFile ( hFile, NULL, NULL, NULL, &IoStatus, devExt->FWBuffer, devExt->FWLength, NULL, NULL );
	
	ntStatus
		= IoStatus.Status;
	if ( !NT_SUCCESS(ntStatus) ) 
	{
		KdPrint((DRVNAME "FAILED to read tcan_usb.hex; ntStatus(%X)\n", ntStatus));
		return ntStatus;
	}
	else
	{
		KdPrint((DRVNAME "ZwReadFile returned %d bytes\n", IoStatus.Information));
	}	
	ZwClose(hFile);

	/* 8051 reset */
    /*  ret =
	usb_control_msg (udev, VENDOR_RQ_OUT, FIRMWARE_LOAD, CPUCS, 0, &cmd,
			 1, 100);
      if (ret < 0)
	{
	  fprintf (stderr, "%s: %s...\n", progname, strerror (errno));
	  exit (1);
	}
	*/

	/*
	 * Transform data into binary format
	 */
	/*
	UINT cpos = 0;
	UINT maxAddr = 0;
	USHORT mlen = 0;
	ULONG len = devExt->FWLength;

	if ( len > 100000 )
	{
		KdPrint((DRVNAME "FAILED: firmware file too big\n"));
		return STATUS_UNSUCCESSFUL;
	}
	/*
	   create a buffer and fill in the data
	*/
	/*
	PCHAR FWBuffer = (PCHAR) ExAllocatePool ( NonPagedPool, len );
	if ( !FWBuffer )
	{
		KdPrint((DRVNAME "FAILED: not enough ressources for firmware binary\n"));
		return STATUS_UNSUCCESSFUL;
	}
	__try
	{
		memset ( FWBuffer, 0, len );
		while ( cpos < len )
		{
			while ( devExt->FWBuffer[cpos] != ':' && cpos < len ) cpos++;
			
			// start of line found, get the address and length
			UINT cc = 0;
			UCHAR cct = 0;
			UINT Addr = 0;
			USHORT Len = 0;
			UINT LenPos = cpos+1;
			UINT AddrPos = cpos+3;

			CHAR c = GetHex ( devExt->FWBuffer[AddrPos] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, a %X, cpos(%X)\n", c, cpos));
				return STATUS_UNSUCCESSFUL;
			}
			Addr |= c;
			Addr <<= 4;
			c = GetHex ( devExt->FWBuffer[AddrPos+1] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, b %X\n", c));
				return STATUS_UNSUCCESSFUL;
			}
			Addr |= c;
			cc = Addr;
			Addr <<= 4;
			c = GetHex ( devExt->FWBuffer[AddrPos+2] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, c %X\n", c));
				return STATUS_UNSUCCESSFUL;
			}
			Addr |= c;
			Addr <<= 4;
			cct = c;
			cct <<= 4;
			c = GetHex ( devExt->FWBuffer[AddrPos+3] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, d %X\n", c));
				return STATUS_UNSUCCESSFUL;
			}
			Addr |= c;
			cct |= c;
			cc += cct;
			// Length
			c = GetHex ( devExt->FWBuffer[LenPos] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, e %X\n", c));
				return STATUS_UNSUCCESSFUL;
			}
			Len |= c;
			Len <<= 4;
			c = GetHex ( devExt->FWBuffer[LenPos+1] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, f %X\n", c));
				return STATUS_UNSUCCESSFUL;
			}
			Len |= c;
			cc += Len;
			if ( Addr > maxAddr )
			{
				maxAddr = Addr; mlen = Len;
			}

			CHAR type = 0;
			c = GetHex ( devExt->FWBuffer[cpos+7] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, g %X\n", c));
				return STATUS_UNSUCCESSFUL;
			}
			type = c;
			type <<= 4;
			c = GetHex ( devExt->FWBuffer[cpos+8] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, h %X\n", c));
				return STATUS_UNSUCCESSFUL;
			}
			type |= c;
			cc += type;

			if ( Addr == 0 && Len == 0 )
			{
				// the type field must be 01 for end of file record				
				if ( type == 0x01 )
				{ break;; }
				else
				{
					KdPrint (( DRVNAME "FAILED: invalid intel hex firmware format, eof record invalid\n" ));
					return STATUS_UNSUCCESSFUL;
				}
			}
			for ( UINT i = 0; i < Len; i++ )
			{
				c = GetHex ( devExt->FWBuffer[cpos+9+(i*2)] );
				if ( c > 0xf )
				{
					KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, g %X\n", c));
					return STATUS_UNSUCCESSFUL;
				}
				FWBuffer[Addr+i] = c;
				FWBuffer[Addr+i] <<= 4;

				c = GetHex ( devExt->FWBuffer[cpos+10+(i*2)] );
				if ( c > 0xf )
				{
					KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, h %X\n", c));
					return STATUS_UNSUCCESSFUL;
				}
				FWBuffer[Addr+i] |= c;
				cc += FWBuffer[Addr+i];
			}
			cc %= 256;
			UCHAR _cc = (UCHAR) cc;
			_cc = ~_cc;
			_cc++;

			c = GetHex ( devExt->FWBuffer[cpos+9+(Len*2)] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, g %X\n", c));
				return STATUS_UNSUCCESSFUL;
			}
			cct = c;
			cct <<= 4;
			c = GetHex ( devExt->FWBuffer[cpos+10+(Len*2)] );
			if ( c > 0xf )
			{
				KdPrint((DRVNAME "FAILED: invalid intel hex firmware format, h %X\n", c));
				return STATUS_UNSUCCESSFUL;
			}
			cct |= c;
			if ( _cc != cct )
			{
				KdPrint((DRVNAME "FAILED: invalid checksum in intel hex firmware format, Addr %X, cc %X, cct %X\n", Addr));
				return STATUS_UNSUCCESSFUL;
			}
			cpos += 8+Len;
		}
	}
	__finally
	{
		if ( FWBuffer )
			ExFreePool ( FWBuffer );
	}*/
	
	/* UINT FWLength = maxAddr+mlen+1;
	 PCHAR buff = (PCHAR) FWBuffer;
	 */
	buff = (PCHAR) devExt->FWBuffer;
	FWLength = devExt->FWLength;

	buffer = 0x1;
	RtlZeroMemory(&urb, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST));
	UsbBuildVendorRequest( 
		&urb,
		URB_FUNCTION_VENDOR_DEVICE,
		sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),
		0,
		0,
		0xA0,									/* .Request (0xB1) VENDOR_RQ_OUT () */
		0x7f92,									/* .Value */
		0,										/* .Index */
		&buffer,								/* .TransferBuffer */
		NULL,
		1,										/* .TransferBufferLength */
		NULL
		);

	ntStatus 
		= Tc_CallToUSB ( deviceObject, &urb );

	if ( !NT_SUCCESS(ntStatus) ) 
	{
		KdPrint (( DRVNAME "FAILED to reset ezusb; ntStatus(%X)\n", ntStatus ));
		return ntStatus;
	}
	else
	{
		KdPrint (( DRVNAME "SUCCESS to reset ezusb\n" ));
	}

	written = 0;
	do
	{
		request = FWLength - written;
		if (request >= 1024)
			request = 1023;
			
		KdPrint (( DRVNAME "writing %d bytes to 0x%04x\n", request, written ));
      
		/*
			ret = usb_control_msg (
				udev,			// device
				VENDOR_RQ_OUT,	// requesttype, 0x40; 0x01000000, host to device, Vendor, to device,
				FIRMWARE_LOAD,	// request 0xA0
				(start + written), // value
				0,				// index
				(buff + written), // data
				request,			// size
				300					// timeout
				);
		*/
		RtlZeroMemory(&urb, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST));
		UsbBuildVendorRequest( 
			&urb,
			URB_FUNCTION_VENDOR_DEVICE,
			sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),
			0,
			0,
			0xA0,									/* .Request (0xB1) VENDOR_RQ_OUT () */
			(USHORT) written,						/* .Value */
			0,										/* .Index */
			(buff + written),						/* .TransferBuffer */
			NULL,
			request,								/* .TransferBufferLength */
			NULL
			);

		ntStatus 
			= Tc_CallToUSB ( deviceObject, &urb );

		if ( !NT_SUCCESS(ntStatus) )
		{
			KdPrint((DRVNAME "FAILED to start can device; ntStatus(%X); UrbStatus(%X)\n", ntStatus, urb.UrbHeader.Status));
			return ntStatus;
		}
		else
			written += request;
	}
	while ( NT_SUCCESS(ntStatus) && written < FWLength );

	  buffer = 0x0;
	  RtlZeroMemory(&urb, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST));
		UsbBuildVendorRequest( 
			&urb,
			URB_FUNCTION_VENDOR_DEVICE,
			sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),
			0,
			0,
			0xA0,									/* .Request (0xB1) VENDOR_RQ_OUT () */
			0x7f92,									/* .Value */
			0,										/* .Index */
			&buffer,								/* .TransferBuffer */
			NULL,
			1,										/* .TransferBufferLength */
			NULL
			);
		ntStatus 
			= Tc_CallToUSB ( deviceObject, &urb );

	if ( !NT_SUCCESS(ntStatus) ) 
	{
		KdPrint (( DRVNAME "FAILED to set ezusb; ntStatus(%X)\n", ntStatus ));
		return ntStatus;
	}
	else
	{
		KdPrint (( DRVNAME "SUCCESS to set ezusb\n" ));
	}

	return STATUS_SUCCESS;
}

CHAR GetHex
( 
 IN CHAR c 
)
{
	/* KdPrint (( DRVNAME "char(%X)\n", c )); */
	if ( c >= '0' && c <= '9' )
		return c-0x30;
	if ( c >= 'A' && c <= 'F' )
		return c-0x37;
	if ( c >= 'a' && c <= 'f' )
		return c-0x57;
	return 0x10;
}