/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description:
 *		Plug and Pray Routines
 *
 * TODO:
 *
 * History of changes:
 *	 2005/03/23 - 0.1 - first version
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "tcan_usb.h"


NTSTATUS Tc_HandleStartDevice
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
);
NTSTATUS Tc_HandleQueryRemoveDevice
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
);
NTSTATUS Tc_HandleRemoveDevice
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
);
NTSTATUS Tc_HandleCancelRequest
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
);
NTSTATUS Tc_HandleQueryStopDevice
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
);
NTSTATUS Tc_HandleStopDevice
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
);
NTSTATUS Tc_HandleSurpriseRemoval
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
);
NTSTATUS Tc_HandleQueryDeviceState
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
);
NTSTATUS Tc_HandleQueryCapabilities
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
);

#ifdef ALLOC_PRAGMA
#pragma alloc_text( PAGE, Tc_DispatchPnp)
#pragma alloc_text( PAGE, Tc_StopQueueAwait)
#pragma alloc_text( PAGE, Tc_StallQueueAwait)
#pragma alloc_text( PAGE, Tc_RestartQueue)
//#pragma alloc_text( PAGE, Tc_SetPnpState)
//#pragma alloc_text( PAGE, Tc_ClearIoQueue)
#pragma alloc_text( PAGE, Tc_HandleStartDevice)
#endif 

/******************************************************************************
 
   Tc_DispatchPnp

 Description:
	Dispatcher routine for pnp irps

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_DispatchPnp
(
 IN PDEVICE_OBJECT	deviceObject, 
 IN PIRP			Irp
)
{
	ULONG				minfc;
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;
	PIO_STACK_LOCATION	ioStack;

	PAGED_CODE();
	
	ntStatus = STATUS_SUCCESS;

	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	if ( devExt->DevState == Removed )
		return Tc_FinishIrp2 ( Irp, STATUS_DELETE_PENDING );

	ntStatus
		= IoAcquireRemoveLock ( &devExt->RemoveLock, Irp );

	if ( !NT_SUCCESS(ntStatus) )
		return Tc_FinishIrp2 ( Irp, ntStatus );

	ioStack 
		= IoGetCurrentIrpStackLocation ( Irp );

	minfc =  ( ioStack->MinorFunction > 24 ? 24 : ioStack->MinorFunction);

	switch ( minfc )
	{
		case IRP_MN_START_DEVICE:
			KdPrint (( DRVNAME "PNP - IRP_MN_START_DEVICE\n" ));
			ntStatus 
				= Tc_HandleStartDevice ( deviceObject, Irp );
			break;

		case IRP_MN_QUERY_STOP_DEVICE:
			KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_STOP_DEVICE\n" ));
			ntStatus
				= Tc_HandleQueryStopDevice ( devExt, Irp );
			IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
			return ntStatus;

		case IRP_MN_CANCEL_STOP_DEVICE:
			KdPrint (( DRVNAME "PNP - IRP_MN_CANCEL_STOP_DEVICE\n" ));
			ntStatus 
				= Tc_HandleCancelRequest ( deviceObject, Irp );
			break;
     
		case IRP_MN_STOP_DEVICE:
			KdPrint (( DRVNAME "PNP - IRP_MN_STOP_DEVICE\n" ));
			ntStatus 
				= Tc_HandleStopDevice ( deviceObject, Irp );
			IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
			return ntStatus;

		case IRP_MN_QUERY_REMOVE_DEVICE:
			KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_REMOVE_DEVICE\n" ));
			ntStatus 
				= Tc_HandleQueryRemoveDevice ( devExt, Irp );
			IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
			return ntStatus;

		case IRP_MN_CANCEL_REMOVE_DEVICE:
			KdPrint (( DRVNAME "PNP - IRP_MN_CANCEL_REMOVE_DEVICE\n" ));
			ntStatus 
				= Tc_HandleCancelRequest ( deviceObject, Irp );
			break;

		case IRP_MN_SURPRISE_REMOVAL:
			KdPrint (( DRVNAME "PNP - IRP_MN_SURPRISE_REMOVAL\n" ));
			ntStatus 
				= Tc_HandleSurpriseRemoval ( deviceObject, Irp );
			IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
			return ntStatus;

		case IRP_MN_REMOVE_DEVICE:
			KdPrint (( DRVNAME "PNP - IRP_MN_REMOVE_DEVICE\n" ));
			ntStatus 
				= Tc_HandleRemoveDevice ( deviceObject, Irp );
			return ntStatus;

		case IRP_MN_QUERY_CAPABILITIES:
			KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_CAPABILITIES\n" ));
			ntStatus 
				= Tc_HandleQueryCapabilities ( deviceObject, Irp );
			break;

		case IRP_MN_QUERY_PNP_DEVICE_STATE:
			KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_PNP_DEVICE_STATE\n" ));
			ntStatus 
				= Tc_HandleQueryDeviceState ( devExt, Irp );
			IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
			return ntStatus;

		default:			
		#if DBG
			switch ( minfc )
			{
				case IRP_MN_QUERY_DEVICE_RELATIONS:
					KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_DEVICE_RELATIONS\n" ));
					break;
				case IRP_MN_QUERY_INTERFACE:
					KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_INTERFACE\n" ));
					break;
				case IRP_MN_QUERY_RESOURCES:
					KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_RESOURCES\n" ));
					break;
				case IRP_MN_QUERY_RESOURCE_REQUIREMENTS:
					KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_RESOURCE_REQUIREMENTS\n" ));
					break;
				case IRP_MN_QUERY_DEVICE_TEXT:
					KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_DEVICE_TEXT\n" ));
					break;
				case IRP_MN_FILTER_RESOURCE_REQUIREMENTS:
					KdPrint (( DRVNAME "PNP - IRP_MN_FILTER_RESOURCE_REQUIREMENTS\n" ));
					break;
				case IRP_MN_READ_CONFIG:
					KdPrint (( DRVNAME "PNP - IRP_MN_READ_CONFIG\n" ));
					break;
				case IRP_MN_WRITE_CONFIG:
					KdPrint (( DRVNAME "PNP - IRP_MN_WRITE_CONFIG\n" ));
					break;
				case IRP_MN_EJECT:
					KdPrint (( DRVNAME "PNP - IRP_MN_EJECT\n" ));
					break;
				case IRP_MN_SET_LOCK:
					KdPrint (( DRVNAME "PNP - IRP_MN_SET_LOCK\n" ));
					break;
				case IRP_MN_QUERY_ID:
					KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_ID\n" ));
					break;
				case IRP_MN_QUERY_BUS_INFORMATION:
					KdPrint (( DRVNAME "PNP - IRP_MN_QUERY_BUS_INFORMATION\n" ));
					break;
				case IRP_MN_DEVICE_USAGE_NOTIFICATION:
					KdPrint (( DRVNAME "PNP - IRP_MN_DEVICE_USAGE_NOTIFICATION\n" ));
					break;
				default:
					KdPrint (( DRVNAME "PNP - UNKNOWN\n" ));
			}
		#endif
			IoSkipCurrentIrpStackLocation ( Irp );
			ntStatus 
				= IoCallDriver ( devExt->NextDeviceObject, Irp );
			IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
			return ntStatus;
	}

	Tc_FinishIrp2 ( Irp, ntStatus );
	IoReleaseRemoveLock ( &devExt->RemoveLock, Irp );
    return ntStatus;
}


/******************************************************************************
 
   Tc_IrpSetEventCR

 Description: 
	This is the completion routine for an Irp. All it has to do is to set
	an event, which the Irps originator waits on.

  @deviceObject - the device Object
  @Irp
  @Context

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_IrpSetEventCR
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp,
 IN PVOID			Context
)
{
	KeSetEvent (
		(PKEVENT) Context, 
		0, 
		FALSE);

	return STATUS_MORE_PROCESSING_REQUIRED;
}


/******************************************************************************
 
   Tc_StallQueueAwait

 Description: 
	Stall the io queue and wait for the active irps to end

  @devExt - the device Extension

 not return value

******************************************************************************/
VOID Tc_StallQueueAwait
(
 IN PDEVICE_EXTENSION	devExt
)
{
	KLOCK_QUEUE_HANDLE hLock;

	KdPrint (( DRVNAME "Tc_StallQueueAwait() at level (%X)\n", KeGetCurrentIrql() ));
	PAGED_CODE ();
	/* 
	   TODO: prio 2
	   Maybe protect this with a spin lock because of multiple instances of this
	   cause race conditions to the events to wait for
    */
	/*
	 * stall the irp queue
	 */
	KeAcquireInStackQueuedSpinLock ( &devExt->IoQueueStateLock, &hLock );
	/*----------------------------------------------------------------*/
		devExt->IoQueueState = Stalled;
	/*----------------------------------------------------------------*/
	KeReleaseInStackQueuedSpinLock ( &hLock );

	/*
	 * wait for all active irps
	 */
	if ( devExt->ReadPending )
	{
		KeClearEvent ( &devExt->ReadStopEvent );
		KeWaitForSingleObject ( 
			&devExt->ReadStopEvent, 
			Executive, 
			KernelMode, 
			FALSE, 
			NULL );
	}

	if ( devExt->WritePending )
	{
		KeClearEvent ( &devExt->WriteStopEvent );
		KeWaitForSingleObject ( 
			&devExt->WriteStopEvent, 
			Executive, 
			KernelMode, 
			FALSE, 
			NULL );
	}
}


/******************************************************************************
 
   Tc_StopQueueAwait

 Description: 
	Stop the io queue and wait for the active irps to end

  @devExt - the device Extension

 not return value

******************************************************************************/
VOID Tc_StopQueueAwait
(
 IN PDEVICE_EXTENSION	devExt
)
{
	KLOCK_QUEUE_HANDLE hLock;

	KdPrint (( DRVNAME "Tc_StopQueueAwait() at level (%X)\n", KeGetCurrentIrql() ));
	PAGED_CODE ();
	/*
	   TODO: prio 2
	   Maybe protect this with a spin lock because of multiple instances of this
	   cause race conditions to the events to wait for
    */
	/*
	   stall the irp queue
	 */
	KeAcquireInStackQueuedSpinLock ( &devExt->IoQueueStateLock, &hLock );
	/*----------------------------------------------------------------*/
		devExt->IoQueueState = InActive;
		devExt->WriteThreadStop = TRUE;
	/*----------------------------------------------------------------*/
	KeReleaseInStackQueuedSpinLock ( &hLock );

	/*
	   Why say FALSE for the wait parameter, and why it crashes when wait equals TRUE and
	   is executed within the spin lock above
	   The kernel use a dispatcher database lock ( w. oney ) for it's thread scheduling
	   things. ReleaseSemaphore, SetEvent, WaitEvent and so on needs to aquire this lock.
	   this means it raise the irql above dispatch_level, don't lower it and the thread's terminating
	   above passive_level at all. Results in a bug check worker_thread_....
    */
	
	KeReleaseSemaphore ( &devExt->WriteThreadSemaphore, 0, 1, FALSE );
	/*
	 *  wait for all active irps
	 */
	KeClearEvent ( &devExt->ReadStopEvent );
	if ( devExt->ReadPending )
	{
		KdPrint (( DRVNAME "Tc_StopQueueAwait: waiting for end of read\n" ));
		
		KeWaitForSingleObject ( 
			&devExt->ReadStopEvent, 
			Executive, 
			KernelMode, 
			FALSE, 
			NULL );
		KdPrint (( DRVNAME "Tc_StopQueueAwait: end of read\n" ));
	}
	
	KeClearEvent ( &devExt->WriteStopEvent );
	if ( devExt->WritePending )
	{
		KdPrint (( DRVNAME "Tc_StopQueueAwait: waiting for end of write\n" ));
		
		KeWaitForSingleObject ( 
			&devExt->WriteStopEvent, 
			Executive, 
			KernelMode, 
			FALSE, 
			NULL );
		KdPrint (( DRVNAME "Tc_StopQueueAwait: end of write\n" ));
	}
	KdPrint (( DRVNAME "Tc_StopQueueAwait: success\n" ));
}


/******************************************************************************
 
   Tc_RestartQueue

 Description: 
	Restart the Queue working routines

  @devExt - the device Extension

 not return value

******************************************************************************/
VOID Tc_RestartQueue
(
 IN PDEVICE_OBJECT	deviceObject
)
{
	KIRQL iLevel;
	PDEVICE_EXTENSION devExt;
	KLOCK_QUEUE_HANDLE hLock;
	KLOCK_QUEUE_HANDLE hLockLock;

	KdPrint (( DRVNAME "Tc_RestartQueue at level (%X) ", KeGetCurrentIrql() ));
	PAGED_CODE ();
	/* 
	   TODO: prio 2
	   Maybe protect this with a spin lock because of multiple instances of this
	   cause race conditions to the events to wait for
     */

	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	/*
	 *  destall the irp queue
	 */
	KeAcquireInStackQueuedSpinLock ( &devExt->IoQueueStateLock, &hLock );
	/*----------------------------------------------------------------*/
		devExt->IoQueueState = Active;
	/*----------------------------------------------------------------*/
	KeReleaseInStackQueuedSpinLock ( &hLock );

	/*
	 *  restart the read irps
	 */
	KeAcquireInStackQueuedSpinLock ( &devExt->ReadLock, &hLock );
	/*----------------------------------------------------------------*/
		if ( !devExt->ReadPending )
		{
			devExt->ReadPending = TRUE;
			/*----------------------------------------------------------------*/
			/* maybe release at passive level as this routine starts because no need for dispatch level */
			iLevel = hLock.OldIrql;
			hLock.OldIrql = DISPATCH_LEVEL;
			KeReleaseInStackQueuedSpinLock ( &hLock );
			/* Restart Read IO */
			KeInsertQueueDpc ( &devExt->IoDpc, NULL, (PVOID) TRUE );
			KeLowerIrql ( iLevel );
		}
		else
		{
			/*----------------------------------------------------------------*/
			KeReleaseInStackQueuedSpinLock ( &hLock );
		}	
	
	/*
	 *  restart the write irps
	 */
	KeAcquireInStackQueuedSpinLock ( &devExt->WriteLock, &hLockLock );
	/*----------------------------------------------------------------*/
		if ( !devExt->WritePending )
		{
			devExt->WritePending = TRUE;
			/*----------------------------------------------------------------*/
			/* iLevel = hLockLock.OldIrql;
			 hLockLock.OldIrql = PASSIVE_LEVEL;
			 KeLowerIrql ( iLevel ); */
			/* Restart the Write IO Thread */
			KeReleaseSemaphore ( &devExt->WriteThreadSemaphore, 0, 1, FALSE );
			KeReleaseInStackQueuedSpinLock ( &hLockLock );
		}
		else
		{
			/*----------------------------------------------------------------*/
			KeReleaseInStackQueuedSpinLock ( &hLockLock );
		}

	KdPrint (( DRVNAME ".. success" ));
}


/******************************************************************************
 
   Tc_SetPnpState

 Description: 
	Helper routine to set the pnp state and previous pnp state

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
VOID Tc_SetPnpState
(
 IN PDEVICE_EXTENSION	devExt,
 IN CANSTATE			pnpState
)
{
	KLOCK_QUEUE_HANDLE hLock;
	KeAcquireInStackQueuedSpinLock ( &devExt->DevStateLock, &hLock );
	/*----------------------------------------------------------------*/
		devExt->DevStatePrev =  devExt->DevState; 
		devExt->DevState = pnpState;
	/*----------------------------------------------------------------*/
	KeReleaseInStackQueuedSpinLock ( &hLock );
}


/******************************************************************************
 
   Tc_ClearIoQueue

 Description: 
	Helper routine to set clear the io queue, means cancel all irps

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
VOID Tc_ClearIoQueue
(
 IN PDEVICE_EXTENSION	devExt
)
{
	QUEUECONTEXT PeekContext;

	KdPrint (( DRVNAME "Tc_ClearIoQueue()\n" ));

	PeekContext.Context = &devExt->NoMatter;
	PeekContext.FileObject = NULL;

	while ( !IsListEmpty ( &devExt->IoQueueHead ) )
	{
		PIRP Irp;
		Irp = IoCsqRemoveNextIrp ( &devExt->IoQueue, &PeekContext );
		Tc_FinishIrp2 ( Irp, STATUS_CANCELLED );
	}

	KdPrint (( DRVNAME ".. succeed\n" ));
}


/******************************************************************************
 
   Tc_HandleStartDevice

 Description: 
	Dispatcher routine for IRP_MN_START_DEVICE

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleStartDevice
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
)
{
    NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;
	KEVENT				Event;

	KdPrint (( DRVNAME "Tc_HandleStartDevice()\n" ));
	PAGED_CODE ();

    devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	Irp->IoStatus.Status = STATUS_SUCCESS;

	/*
	  Pass it down and wait for the results
	  --------------------------------------*/
    KeInitializeEvent ( &Event, NotificationEvent, FALSE );
    IoCopyCurrentIrpStackLocationToNext ( Irp );
    IoSetCompletionRoutine (
		Irp, 
        (PIO_COMPLETION_ROUTINE) Tc_IrpSetEventCR, 
        (PVOID)&Event, 
        TRUE, 
        TRUE, 
        TRUE);

	if( IoCallDriver ( devExt->NextDeviceObject, Irp ) == STATUS_PENDING ) 
        KeWaitForSingleObject (
			&Event, 
            Executive, 
            KernelMode, 
            FALSE, 
            NULL );
	/*
	  ...
	  --------------------------------------*/
    ntStatus = Irp->IoStatus.Status;

    if( !NT_SUCCESS(ntStatus) ) 
	{
        KdPrint (( DRVNAME "Lower drivers failed this Irp\n" ));
        return ntStatus;
    }

	if ( devExt->PhysicalReconnectNeeded )
		return STATUS_UNSUCCESSFUL;

    ntStatus = Tc_StartDevice ( deviceObject );

	Tc_SetPnpState ( devExt, Working );

    return ntStatus;
}


/******************************************************************************
 
   Tc_HandleQueryStopDevice

 Description: 
	Dispatcher routine for IRP_MN_QUERY_STOP_DEVICE
	Possibly next states are: Working or Surpriseremoved

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleQueryStopDevice
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
)
{
	KdPrint (( DRVNAME "Tc_HandleQueryStopDevice()\n" ));
	
	Irp->IoStatus.Status = STATUS_SUCCESS;
	
	if ( devExt->DevState == Working || devExt->PhysicalReconnectNeeded )
	{
		KdPrint (( DRVNAME "QUERY_STOPDEV: stall queue await\n" ));
		Tc_StallQueueAwait ( devExt );
		KdPrint (( DRVNAME "QUERY_STOPDEV: Pending Stop\n" ));
		Tc_SetPnpState ( devExt, PendingStop );
	}

	KdPrint (( DRVNAME "QUERY_STOPDEV: success\n" ));
	IoSkipCurrentIrpStackLocation ( Irp );
	return IoCallDriver ( devExt->NextDeviceObject, Irp );
}


/******************************************************************************
 
   Tc_HandleStopDevice

 Description: 
	Dispatcher routine for IRP_MN_STOP_DEVICE

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleStopDevice
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
)
{
	PDEVICE_EXTENSION devExt;

	KdPrint (( DRVNAME "Tc_HandleStopDevice()\n" ));
	
	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	Irp->IoStatus.Status = STATUS_SUCCESS;

	/*
	  c: stop the can device ( inactivate the io-queue, polling irps, free RX-Buffer )
	  ---------------------------------------------------------------------------------*/
	Tc_StopCan ( deviceObject, TRUE);

	/*
	  d: free the irp resources, and submit the device a reset configuration descriptor
	  also free the memory for the configuration descriptor
	  ---------------------------------------------------------------------------------*/	
	Tc_StopDevice ( deviceObject, TRUE );

	Tc_SetPnpState ( devExt, Stopped );
	
	IoSkipCurrentIrpStackLocation ( Irp );
	return IoCallDriver ( devExt->NextDeviceObject, Irp );
}


/******************************************************************************
 
   HandleQueryRemove

 Description: 
	Dispatcher routine for IRP_MN_QUERY_REMOVE_DEVICE.
	Plug and Play manager wants us to turn to PendingRemove state.
	Possibly previous states are: Working
	Possibly next states are: Removed, Surpriseremoved or Working

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleQueryRemoveDevice
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
)
{
	KdPrint (( DRVNAME "Tc_HandleQueryRemoveDevice " ));
	Irp->IoStatus.Status = STATUS_SUCCESS;

	if ( devExt->DevState == Working )
	{
		/*
			special handling for windows 98 is necessary for later implementations
			this driver supports only Windows XP
			---------------------------------------------------------------------------------*/	
		Tc_StallQueueAwait ( devExt );
	}

	Tc_SetPnpState ( devExt, PendingRemove );

	KdPrint (( DRVNAME ".. success" ));
	IoSkipCurrentIrpStackLocation ( Irp );
	return IoCallDriver ( devExt->NextDeviceObject, Irp );
}


/******************************************************************************
 
   HandleQueryDeviceState

 Description: 
	Set the Flags if device is function improperly otherwise pass it down

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleQueryDeviceState
(
 IN PDEVICE_EXTENSION	devExt,
 IN PIRP				Irp
)
{
	KdPrint (( DRVNAME "Tc_HandleQueryDeviceState\n" ));
	
	if ( devExt->PhysicalReconnectNeeded )
	{		
        PPNP_DEVICE_STATE dstate;

		KdPrint (( DRVNAME "PNP_DEVSTATE: change device state !\n" ));
        dstate = (PPNP_DEVICE_STATE) &Irp->IoStatus.Information;
        *dstate |= PNP_DEVICE_FAILED;
        *dstate |= PNP_DEVICE_DISABLED;
		Irp->IoStatus.Status = STATUS_SUCCESS;

		Tc_SetPnpState ( devExt, Stopped );
	}

	IoCopyCurrentIrpStackLocationToNext ( Irp );
	return IoCallDriver ( devExt->NextDeviceObject, Irp );
}


/******************************************************************************
 
   TC_HandleCancelRequest

 Description: 
	Dispatcher routine for IRP_MN_CANCEL_REMOVE_DEVICE
	or IRP_MN_CANCEL_STOP_DEVICE
	Plug and Play manager wants us to cancel the remove or stop request
	and turn to Working state.
	Previous state is PendingRemove or PendingStop
	Next state is Working

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleCancelRequest
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
)
{
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;
	KEVENT				Event;

	KdPrint (( DRVNAME "Tc_HandleCancelRequest (Cancel remove or stop)\n" ));

	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	Irp->IoStatus.Status = STATUS_SUCCESS;

	if ( devExt->DevState == PendingRemove || devExt->DevState == PendingStop)
	{
		/*
		   Pass it down and wait for the results
		   --------------------------------------*/
		KeInitializeEvent ( &Event, NotificationEvent, FALSE );
		IoCopyCurrentIrpStackLocationToNext ( Irp );
		IoSetCompletionRoutine (
			Irp, 
			(PIO_COMPLETION_ROUTINE) Tc_IrpSetEventCR, 
			(PVOID)&Event, 
			TRUE, 
			TRUE, 
			TRUE);

		if( IoCallDriver ( devExt->NextDeviceObject, Irp ) == STATUS_PENDING ) 
			KeWaitForSingleObject (
				&Event, 
				Executive, 
				KernelMode, 
				FALSE, 
				NULL );
		/*
		   ...
		   --------------------------------------*/
		ntStatus = Irp->IoStatus.Status;

		if (NT_SUCCESS(ntStatus))
			Tc_RestartQueue ( deviceObject );
		
		/* maybe not neccessary or only for pending stop */
		Tc_SetPnpState ( devExt, Working );

		return ntStatus;
	}
	
	IoSkipCurrentIrpStackLocation ( Irp );
	return IoCallDriver ( devExt->NextDeviceObject, Irp );
}


/******************************************************************************
 
   Tc_HandleRemoveDevice

 Description: 
	Dispatcher routine for IRP_MN_REMOVE_DEVICE
	Plug and Play manager wants us to make things so it can remove the device
	Previous state is PendingRemove ( or maybe working ?)
	Next state is Removed

 Note:
	Win98/ME does not send a IRP_MN_SURPRISE_REMOVAL request, hence a later
	adaption of this driver for this operating system needs to indicate a
	surprise removal, if it come out of the blue ( as walter oney described
	in his book ).
	Win98 (original) also does not handle remove locks, which where used
	in this driver.
	Win98SE/ME support it but remove lock and wait.
	To handle this situation should not be difficult, if using wrapper
	fuctions, but this is not part of the current diploma thesis.

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleRemoveDevice
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
)
{
	NTSTATUS			ntStatus;
	PDEVICE_EXTENSION	devExt;

	KdPrint (( DRVNAME "Tc_HandleRemoveDevice()\n" ));

	Irp->IoStatus.Status = STATUS_SUCCESS;

	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;
	
	/*int i = 0;
	while ( i < 3 ){*/
	if ( devExt->DevState != SurpriseRemoved )
	{
		/*
		   c: stop the can device ( inactivate the io-queue, polling irps, free RX-Buffer )
		   --------------------------------------------------------------------------------*/
		if ( devExt->CanStarted ) Tc_StopCan ( deviceObject, TRUE );

		/*
		   d: free the irp resources, and submit the device a reset configuration descriptor
		   also free the memory for the configuration descriptor
		   ( devExt->DevState == SurpriseRemoved ? FALSE : TRUE ) 
		   ---------------------------------------------------------------------------------*/
		Tc_StopDevice ( deviceObject, TRUE);
	}

	Tc_SetPnpState ( devExt, Removed );
	/*i++;
}*/
	IoSkipCurrentIrpStackLocation(Irp);

	ntStatus 
		= IoCallDriver ( devExt->NextDeviceObject, Irp );

	IoReleaseRemoveLockAndWait ( &devExt->RemoveLock, Irp );
	
	KdPrint (( DRVNAME "about to remove device from stack..\n" ));
	
	Tc_RemoveDevice ( deviceObject );

	KdPrint (( DRVNAME ".. removed !\n" ));
	return ntStatus;
}


/******************************************************************************
 
   Tc_HandleSurpriseRemoval

 Description: 
	Dispatcher routine for IRP_MN_SURPRISE_REMOVAL.
	Plug and Play manager wants us to recognize a surprise removal
	Possibly previous: Working, Stopped, PendingRemove, PendingStop
	Possibly next state is Removed

  @devExt - the device Extension
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleSurpriseRemoval
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
)
{
	PDEVICE_EXTENSION devExt;
	/*int i = 0; */

	KdPrint (( DRVNAME "Tc_HandleSurpriseRemoval()\n" ));
	Irp->IoStatus.Status = STATUS_SUCCESS;

	devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

	/*i = 0;
	while ( i < 3 )
	{*/
	/*
		c: stop the can device ( deactivate the io-queue, polling irps, free RX-Buffer )
		--------------------------------------------------------------------------------*/
	if ( devExt->CanStarted ) Tc_StopCan ( deviceObject, FALSE );

	/*
		d: free the irp resources, and submit the device a reset configuration descriptor
		also free the memory for the configuration descriptor
		----------------------------------------------------------------------------------*/
	Tc_StopDevice ( deviceObject, FALSE );

	Tc_SetPnpState ( devExt, SurpriseRemoved );
	/*i++;
	}*/

	IoSkipCurrentIrpStackLocation ( Irp );
	return IoCallDriver ( devExt->NextDeviceObject, Irp );
}


/******************************************************************************
 
   Tc_HandleQueryCapabilities

 Description: 
	routine for IRP_MN_QUERY_CAPABILITIES.
	Plug and Play manager wants us to supply the device capabilities

  @deviceObject - the device object
  @Irp

 returns a ntstatus

******************************************************************************/
NTSTATUS Tc_HandleQueryCapabilities
(
 IN PDEVICE_OBJECT	deviceObject,
 IN PIRP			Irp
)
{
	NTSTATUS				ntStatus;
	PDEVICE_EXTENSION		devExt;
	PIO_STACK_LOCATION		ioStack;
	PDEVICE_CAPABILITIES	devCap;
	KEVENT					Event;

	KdPrint (( DRVNAME "Tc_HandleQueryCapabilities\n" ));

    devExt
		= (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

    ioStack
		= IoGetCurrentIrpStackLocation ( Irp );

    devCap
		 = ioStack->Parameters.DeviceCapabilities.Capabilities;

    if ( devCap->Version < 1 || devCap->Size < sizeof(DEVICE_CAPABILITIES) )
	{
        KdPrint (( DRVNAME "Version < 1 or not a valid structure\n" ));
        return STATUS_UNSUCCESSFUL;
    }

    /*
			DDK says add capabilities here
       Add the SurpriseRemovalOK flag before passing it down.
    */
	devCap->Removable = TRUE;
    devCap->SurpriseRemovalOK = TRUE;
	devCap->WakeFromD0 = TRUE;
	devCap->WakeFromD1 = FALSE;
	devCap->WakeFromD2 = FALSE;
	devCap->WakeFromD3 = TRUE;
	devCap->DeviceWake = PowerDeviceD3;
	/*
	   according wdm_pm11.pdf, P.10, a device not supporting wake up
	   should have these values
	*/
	devCap->DeviceState[PowerSystemWorking] = PowerDeviceD0;
	devCap->DeviceState[PowerSystemSleeping1] = PowerDeviceD3;
	devCap->DeviceState[PowerSystemSleeping2] = PowerDeviceD3;
	devCap->DeviceState[PowerSystemSleeping3] = PowerDeviceD3;
	devCap->DeviceState[PowerSystemHibernate] = PowerDeviceD3;
	devCap->DeviceState[PowerSystemShutdown] = PowerDeviceD3;
	/*
	devCap->DeviceD1 = FALSE;
	devCap->DeviceD2 = FALSE;
	devCap->LockSupported = FALSE;
	devCap->EjectSupported = FALSE;
	devCap->Removable = TRUE;
	devCap->DockDevice = FALSE;
	devCap->UniqueID = FALSE;
	devCap->SilentInstall = FALSE;
	devCap->RawDeviceOK = FALSE;
	devCap->SurpriseRemovalOK = TRUE;
	devCap->WakeFromD0 = FALSE; //-->1
	devCap->WakeFromD1 = FALSE;
	devCap->WakeFromD2 = FALSE;
	devCap->WakeFromD3 = FALSE; //-->1
	devCap->HardwareDisabled = FALSE;
	devCap->NonDynamic = FALSE;
	devCap->WarmEjectSupported = FALSE;
	devCap->SystemWake = 2;
	devCap->DeviceWake = 1; //-->4
	devCap->D1Latency = 0;
	devCap->D2Latency = 0;
	devCap->D3Latency = 0;
	devCap->DeviceState[PowerSystemUnspecified] = 0;
	devCap->DeviceState[PowerSystemWorking] = 1;
	devCap->DeviceState[PowerSystemSleeping1] = 4;
	devCap->DeviceState[PowerSystemSleeping2] = 4;
	devCap->DeviceState[PowerSystemSleeping3] = 4;
	devCap->DeviceState[PowerSystemHibernate] = 4;
	devCap->DeviceState[PowerSystemShutdown] = 4;
	devCap->DeviceState[PowerSystemMaximum] = 2;
	*/
    Irp->IoStatus.Status = STATUS_SUCCESS;

	/*
	  Pass it down and wait for the results
	  --------------------------------------*/
	KeInitializeEvent ( &Event, NotificationEvent, FALSE );
	IoCopyCurrentIrpStackLocationToNext ( Irp );
	IoSetCompletionRoutine (
		Irp, 
		(PIO_COMPLETION_ROUTINE) Tc_IrpSetEventCR, 
		(PVOID)&Event, 
		TRUE, 
		TRUE, 
		TRUE);

	if( IoCallDriver ( devExt->NextDeviceObject, Irp ) == STATUS_PENDING ) 
		KeWaitForSingleObject (
			&Event, 
			Executive, 
			KernelMode, 
			FALSE, 
			NULL );

	ntStatus = Irp->IoStatus.Status;

	/*
	  DDK says remove capabilities here
	  --------------------------------------*/

	devCap->SurpriseRemovalOK = TRUE;

	RtlCopyMemory ( &devExt->CanCapabilities, devCap, sizeof(DEVICE_CAPABILITIES));

	/*
	KdPrint (( DRVNAME "PWR: Version %i\n", devCap->Version));
	KdPrint (( DRVNAME "PWR: DeviceD1 %i\n", devCap->DeviceD1));
	KdPrint (( DRVNAME "PWR: DeviceD2 %i\n", devCap->DeviceD2));
	KdPrint (( DRVNAME "PWR: LockSupported %i\n", devCap->LockSupported));
	KdPrint (( DRVNAME "PWR: EjectSupported %i\n", devCap->EjectSupported));
	KdPrint (( DRVNAME "PWR: Removable %i\n", devCap->Removable));
	KdPrint (( DRVNAME "PWR: DockDevice %i\n", devCap->DockDevice));
	KdPrint (( DRVNAME "PWR: UniqueID %i\n", devCap->UniqueID));
	KdPrint (( DRVNAME "PWR: SilentInstall %i\n", devCap->SilentInstall));
	KdPrint (( DRVNAME "PWR: RawDeviceOK %i\n", devCap->RawDeviceOK));
	KdPrint (( DRVNAME "PWR: SurpriseRemovalOK %i\n", devCap->SurpriseRemovalOK));
	KdPrint (( DRVNAME "PWR: WakeFromD0 %i\n", devCap->WakeFromD0));
	KdPrint (( DRVNAME "PWR: WakeFromD1 %i\n", devCap->WakeFromD1));
	KdPrint (( DRVNAME "PWR: WakeFromD2 %i\n", devCap->WakeFromD2));
	KdPrint (( DRVNAME "PWR: WakeFromD3 %i\n", devCap->WakeFromD3));
	KdPrint (( DRVNAME "PWR: HardwareDisabled %i\n", devCap->HardwareDisabled));
	KdPrint (( DRVNAME "PWR: NonDynamic %i\n", devCap->NonDynamic));
	KdPrint (( DRVNAME "PWR: WarmEjectSupported %i\n", devCap->WarmEjectSupported));
	KdPrint (( DRVNAME "PWR: SystemWake %i\n", devCap->SystemWake));
	KdPrint (( DRVNAME "PWR: DeviceWake %i\n", devCap->DeviceWake));
	KdPrint (( DRVNAME "PWR: D1Latency %i\n", devCap->D1Latency));
	KdPrint (( DRVNAME "PWR: D2Latency %i\n", devCap->D2Latency));
	KdPrint (( DRVNAME "PWR: D3Latency %i\n", devCap->D3Latency));
	KdPrint (( DRVNAME "PWR: DeviceState / PowerSystemUnspecified %i\n", devCap->DeviceState[PowerSystemUnspecified]));
	KdPrint (( DRVNAME "PWR: DeviceState / PowerSystemWorking %i\n", devCap->DeviceState[PowerSystemWorking]));
	KdPrint (( DRVNAME "PWR: DeviceState / PowerSystemSleeping1 %i\n", devCap->DeviceState[PowerSystemSleeping1]));
	KdPrint (( DRVNAME "PWR: DeviceState / PowerSystemSleeping2 %i\n", devCap->DeviceState[PowerSystemSleeping2]));
	KdPrint (( DRVNAME "PWR: DeviceState / PowerSystemSleeping3 %i\n", devCap->DeviceState[PowerSystemSleeping3]));
	KdPrint (( DRVNAME "PWR: DeviceState / PowerSystemHibernate %i\n", devCap->DeviceState[PowerSystemHibernate]));
	KdPrint (( DRVNAME "PWR: DeviceState / PowerSystemShutdown %i\n", devCap->DeviceState[PowerSystemShutdown]));
	KdPrint (( DRVNAME "PWR: DeviceState / PowerSystemMaximum %i\n", devCap->DeviceState[PowerSystemMaximum]));


	PowerSystemUnspecified = 0,
    PowerSystemWorking     = 1,
    PowerSystemSleeping1   = 2,
    PowerSystemSleeping2   = 3,
    PowerSystemSleeping3   = 4,
    PowerSystemHibernate   = 5,
    PowerSystemShutdown    = 6,
    PowerSystemMaximum     = 7
	*/
    return ntStatus;
}