/*
 * USB-Tiny-CAN Driver
 *
 *
 * Copyright (C) 2005 Chi Tai Dang, thesis at chi-tai dot info
 *						All rights reserved
 *
 * Description:
 *		global definitions and api values
 *		outsourced because these values could also be used by user applications
 *
 * TODO:
 *
 * History of changes:
 *	 2005/03/15 - 0.1 - first version
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef TCAN_USB_GLOBAL_H
#define TCAN_USB_GLOBAL_H


/* frame information register*/
#define CAN_SFF						0x00	/* standard frame bit */
#define CAN_EFF						0x80	/* extended frame bit */
#define CAN_RTR						0x40	/* remote transmition request bit */
#define CAN_DLC						0x0F	/* data length code */

/* mode register*/
#define CAN_RESET_MODE_REQUEST		0x01
#define CAN_LISTEN_ONLY_MODE		0x02
#define CAN_SELFTEST_MODE			0x04
#define CAN_ACCEPTANCE_FILTER_MODE	0x08
#define CAN_SLEEP_MODE				0x10

typedef
enum baud_index {
	CAN_BAUD_10K = 1,
	CAN_BAUD_20K,
	CAN_BAUD_50K,
	CAN_BAUD_100K,
	CAN_BAUD_125K,
	CAN_BAUD_250K,
	CAN_BAUD_500K,
	CAN_BAUD_800K,
	CAN_BAUD_1000K,
	CAN_BAUD_ENDVALUE
} CANBAUD, *PCANBAUD;

typedef unsigned int UINT;
/*
   max CAN message size
*/
#define CAN_MSG_LENGTH				8

/*
   That's the can message structure
*/
typedef
struct _canmsg 
{
	ULONG			id;
	//(std/ext,rtr,dlc) as in Frame Info Reg
	UINT			flags;	
    UCHAR			msgdata[CAN_MSG_LENGTH];
	// timestamp for received msg
	LARGE_INTEGER	rtime;	
}
CANMSG, *PCANMSG;

/*
   Modes for IOCTL_SET_IO_MODE
*/
typedef
enum _IoMode
{
	PollMode,
	EventMode
} 
CIOMODE, *PCIOMODE;

/*
   structure of the interrupt data received by IOCTL_GET_INTERRUPT
*/
typedef
struct uci_err_status 
{
   UCHAR st_code;
   UCHAR errors;
   UCHAR overruns;
   UCHAR irq;
   UCHAR status;
   UCHAR err_code_cap;
   UCHAR err_warn_lim;
   UCHAR rx_err_cnt;
   UCHAR tx_err_cnt;
   UCHAR fw_fifo_ovr;
   UCHAR mode;
}
UCI_INT_DATA, *PUCI_INT_DATA;

#endif
